package de.fraunhofer.iem.dataprovider.toolRun

import de.fraunhofer.iem.dataprovider.gitlab.Platform
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.taskManager.tasks.GitRepository
import de.fraunhofer.iem.dataprovider.taskManager.tasks.RepositoryDetailsDto
import de.fraunhofer.iem.dataprovider.taskManager.tasks.toDbObject
import de.fraunhofer.iem.dataprovider.toolRun.model.*
import org.springframework.stereotype.Service
import java.util.*

@Service
class ToolRunService(
    private val repositoryRepository: RepositoryRepository,
    private val repositoryDetailsRepository: RepositoryDetailsRepository,
    private val toolRunRepository: ToolRunRepository,
    private val toolRepository: ToolRepository,
    private val ruleRepository: RuleRepository
) {

    private val logger = getLogger(javaClass)
    fun getLatestToolRunsForRepo(repoId: Long, platform: Platform): List<ToolRun> {
        val repo = repositoryRepository.findByRepoIdAndPlatform(repoId, platform)
        if (repo !== null) {
            return getLatestToolResults(repo)
        } else {
            logger.warn("Requested repository doesn't exist.")
        }
        return emptyList()
    }

    fun getLatestToolRunsForRepo(repoId: UUID): List<ToolRun> {
        val repo = repositoryRepository.findById(repoId)
        if (repo.isPresent) {
            return getLatestToolResults(repo.get())
        }
        return emptyList()
    }

    private fun getLatestToolResults(repo: Repository): List<ToolRun> {
        return repo.tools.map { tool ->
            toolRunRepository.findFirstByToolIdAndRepositoryIdOrderByTimeStampDesc(tool.id!!, repo.id!!)
        }

    }

    // TODO: this func should be a suspend fun. We can achieve this by
    //  wrapping the db return types in flux, mono, or futures
    fun saveSarif(sarif: Sarif, repoId: UUID) {

        val repo = repositoryRepository.findById(repoId)

        sarif.runs.forEach { run ->

            val tool = findOrCreateTool(run.tool)
            val rules = findRulesForTool(tool.id)

            createNewRulesAndSaveToTool(run.tool.driver.rules, rules, tool)


            val toolResults = run.results.map { sarifResult ->
                val tr = sarifResult.asDbObject()
                // link rules to results
                rules.filter { it.sarifRuleId == sarifResult.ruleId }
                    .forEach { it.addResultLocation(tr) }
                tr
            }


            val toolRun = run.asDbObject()
            toolRun.tool = tool
            toolRun.addRules(rules)
            toolRun.addFindings(toolResults)

            val result = toolRunRepository.save(toolRun)
            toolRepository.save(tool)

            repo.ifPresent {
                if (!it.tools.contains(tool)) {
                    it.tools.add(tool)
                }
                it.addToolResult(result)
                repositoryRepository.save(it)
            }
        }
    }

    private fun createNewRulesAndSaveToTool(
        sarifRules: List<SarifRule>,
        rules: MutableList<Rule>,
        tool: Tool
    ) {
        sarifRules.forEach { sarifRule ->
            if (rules.find { it.sarifRuleId == sarifRule.id } == null) {
                val rule = ruleRepository.save(sarifRule.asDbObject())
                rules.add(rule)
                tool.addRule(rule)
            }
        }
    }

    private fun findOrCreateTool(tool: SarifTool): Tool {
        val fullName = tool.driver.fullName
        val name = tool.driver.name
        val version = tool.driver.version

        return toolRepository.findByFullNameIgnoreCaseAndNameIgnoreCaseAndVersionIgnoreCase(fullName, name, version)
            ?: toolRepository.save(tool.asDbObject())
    }

    private fun findRulesForTool(id: UUID?): MutableList<Rule> {
        return if (id != null) {
            ruleRepository.findByTool_Id(id).toMutableList()
        } else {
            mutableListOf()
        }
    }

    fun getRepoOrCreate(gitRepository: GitRepository): Repository {
        var repo = repositoryRepository.findByRepoIdAndPlatform(gitRepository.id, gitRepository.platform)
        if (repo == null) {
            repo = repositoryRepository.save(gitRepository.toDbObject())
        }
        return repo
    }

    // TODO: We should decide if we want to return optionals (null if not found) or throw exceptions or ...
    // TODO: Must be consistent across all methods
    fun findRepoByID(id: UUID): Repository? {
        val repository = repositoryRepository.findById(id)
        if (repository.isEmpty) {
            return null
        }
        return repository.get()
    }

    fun createRepositoryDetails(repositoryDetailsDto: RepositoryDetailsDto): RepositoryDetails {
        return repositoryDetailsRepository.save(repositoryDetailsDto.toDbObject())
    }

    fun getLatestRepositoryDetailsByRepositoryId(repositoryId: UUID): RepositoryDetails? {
        val repositoryDetails = repositoryDetailsRepository.findFirstByRepository_IdOrderByTimestampDesc(repositoryId);
        if (repositoryDetails.isEmpty) {
            return null
        }
        return repositoryDetails.get();
    }

    fun getRepositoryDetailsById(id: UUID): RepositoryDetails? {
        val repositoryDetails = repositoryDetailsRepository.findById(id);
        if (repositoryDetails.isEmpty) {
            return null
        }
        return repositoryDetails.get()
    }

    fun getAllRepositories(): List<Repository> {
        return repositoryRepository.findAll()
    }
}
